@@ -105,11 +105,11 @@ func abort() {
 }
 
 //export write
-func write(fd int32, buf *byte, count int) int {
+func libc_write(fd int32, buf *byte, count int) int {
 	// TODO: Proper handling write
 	for i := 0; i < count; i++ {
 		putchar(*buf)
-		buf = (*byte)(unsafe.Pointer(uintptr(unsafe.Pointer(buf)) + 1))
+		buf = (*byte)(unsafe.Add(unsafe.Pointer(buf), 1))
 	}
 	return count
 }
